# file: fact.py
# author: Bob Muller
#
# CS1101 Computer Science I
#
# This file contains three versions of the factorial function. All of them
# illustrate the use of recursion to express a repetitive process. The first
# version, fact1, follows the mathematical definition but is inefficient.
#
# fact1 : int -> int
#
def fact1(n):
  if n == 0:
    return 1
  else:
    return n * fact1(n - 1)
  
# fact2 : int -> int
#
# The call fact2(n) computes n!. Unlike fact1, it computes the partial
# answer "on the way in" rather than "on the way out". In other words,
# fact2 is TAIL RECURSIVE. Most modern programming languages implement
# tail recursion efficiently. (Unfortunately, Python doesn't.) Also, note
# that fact2 defines a nested helper function.
#
def fact2(n):
  def repeat(n, acc):
    if n == 0:
      return acc
    else:
      return repeat(n - 1, n * acc)
  return repeat(n, 1)

# fact3 : int -> int
#
# The call fact3(n) computes n!. It follows the more efficient strategy
# of fact2 but it also uses Python's assert form to avoid an infinite loop
# if someone called it with a negative integer.
#
def fact3(n):
  assert n >= 0
  def repeat(n, acc):
    if n == 0:
      return acc
    else:
      return repeat(n - 1, n * acc)
  return repeat(n, 1)

